"""
    This package allows you to measure the power drained by your
    computer or server during the execution of a function
"""
__all__ = ["PowerMeter", "MagicPowerMeter"]

from .magic_power_meter import MagicPowerMeter
from .power_meter import PowerMeter



import os
from pathlib import Path

from fuzzywuzzy import fuzz  # type: ignore

PACKAGE_PATH = Path(os.path.dirname(os.path.abspath(__file__)))
HOME_DIR = Path.home()

ENERGY_MIX_DATABASE = Path("data/ademe_energy_mix_by_country.csv")
ENERGY_MIX_COLUMN = "Energy mix (kgCO2/kWh)"
COUNTRY_CODE_COLUMN = "ISO"
COUNTRY_NAME_COLUMN = "Country"

POWERLOG_PATH_LINUX = Path("/sys/class/powercap/intel-rapl")
MSR_PATH_LINUX_TEST = Path("/dev/cpu/0/msr")
MAC_INTELPOWERLOG_FILENAME = "intelPowerLog.csv"
WIN_INTELPOWERLOG_FILENAME = "PwrData_*.csv"

LOGGING_FILE = "power_logs.csv"

TOTAL_CPU_TIME = "Total Elapsed CPU Time (sec)"
TOTAL_GPU_TIME = "Total Elapsed GPU Time (sec)"
TOTAL_ENERGY_ALL = "Cumulative Package Energy (mWh)"
TOTAL_ENERGY_CPU = "Cumulative IA Energy (mWh)"
TOTAL_ENERGY_PROCESS_CPU = "Cumulative process CPU Energy (mWh)"
TOTAL_ENERGY_GPU = "Cumulative GPU Energy (mWh)"
TOTAL_ENERGY_MEMORY = "Cumulative DRAM Energy (mWh)"
TOTAL_ENERGY_PROCESS_MEMORY = "Cumulative process DRAM Energy (mWh)"
CPU_PERCENT_USAGE = "CPU_percent_usage"
MEMORY_PERCENT_USAGE = "memory_percent_usage"

AVAILABLE_STEPS = [
    "inference",
    "training",
    "other",
    "test",
    "run",
    "preprocessing",
]


def match(s, options, threshold=80):
    """Fuzzy matching function."""

    results = (
        (opt, fuzz.token_set_ratio(opt.lower(), s.lower())) for opt in options
    )
    results = sorted(
        [(opt, v) for opt, v in results if v >= threshold],
        key=lambda tupl: tupl[1],
        reverse=True,
    )
    if results:
        return results[0][0]

    return s


def normalize(normalize_s, default_value=""):
    """Normalization function.

    Args:
        normalize_s (str): string to normalize.
        default_value (str, optional): [description]. Defaults to "".

    Returns:
        [type]: [description]
    """
    return str(normalize_s).lower() if normalize_s else default_value.lower()



"""
Wrapper of the Python class PowerMeter for a notebook usage
"""
__all__ = ["MagicPowerMeter"]

from IPython.core.magic import Magics, cell_magic, magics_class  # type: ignore
from IPython.core.magic_arguments import (  # type: ignore
    argument,
    magic_arguments,
    parse_argstring,
)

from .power_meter import PowerMeter


@magics_class
class MagicPowerMeter(Magics):
    """
    This class will be used to use a PowerMeter instance in a notebook cell.

    This class aims at allowing the usage of ipython magic functions.
    It is not made to be used alone

    Parameters
    ----------
    Magics : IPythonShell
        An ipython shell

    See Also
    --------
    MagicPowerMeter.measure_power : Measure the power consumption using a \
        ipython magic function
    PowerMeter : Instanciate a PowerMeter

    Examples
    --------

    Load the ``MagicPowerMeter`` extension then declare a PowerMeter as usual

    .. code-block:: python

        %load_ext carbonai.MagicPowerMeter
        from carbonai import PowerMeter
        power_meter = PowerMeter(project_name="MNIST classifier", \
            is_online=False, location="FR")

    """

    def __init__(self, shell):
        super().__init__(shell)
        self.power_meter = None

    @magic_arguments()
    @argument("power_meter", help="The PowerMeter object of this project")
    @argument("package", type=str, help="The name of the package used here")
    @argument("algorithm", type=str, help="The algorithm type used here")
    @argument("--step", type=str, help="Type of data used")
    @argument("--data_type", help="Type of data used")
    @argument("--data_shape", help="Shape of the data used")
    @argument(
        "--algorithm_params",
        help="Some informative parameters \
            used in your algorithm",
    )
    @argument("--comments", type=str, help="Comments to describe what is done")
    @cell_magic
    def measure_power(self, line, cell):
        """
        An IPython magic function to measure the power consumption of \
        a given cell

        Parameters
        ----------
        power_meter : carbonai.PowerMeter
            A PowerMeter object used to collect the carbon logs
        package : str
            A string describing the package used by this function \
                (e.g. sklearn, Pytorch, ...)
        algorithm : str
            A string describing the algorithm used in the function \
                monitored (e.g. RandomForestClassifier, \
                ResNet121, ...)
        step : {'inference', 'training', 'other', 'test', \
            'run', 'preprocessing'}, optional
            A string to provide useful information on the current stage of \
                the alogrithm
        data_type : \
            {'tabular', 'image', 'text', 'time series', 'other'}, optional
            A string describing the type of data used for training
        data_shape : str or tuple, optional
            A string or tuple describing the quantity of data used
        algorithm_params : str, optional
            A string describing the parameters used by the algorithm
        comments : str, optional
            A string to provide any useful information

        Returns
        -------

        See also
        --------
        MagicPowerMeter : Loads the jupyter carbonai extension
        PowerMeter : Instantiate a PowerMeter
        PowerMeter.measure_power : Another way to measure the power \
            usage of some code

        Examples
        --------
        Load the ``MagicPowerMeter`` extension then declare a PowerMeter \
            as usual

        .. code-block:: python

            %load_ext carbonai.MagicPowerMeter
            from carbonai import PowerMeter
            power_meter = PowerMeter(project_name="MNIST classifier", \
                is_online=False, location="FR")

        In each cell you want to measure, you can then use the \
            ``measure_power`` cell magic

        .. code-block:: python

            %%measure_power power_meter "package_name_used" \
                "algorithm" --step "training" --data_type "tabular" \
                --data_shape "your_data_shape" \
                --algorithm_params "n_estimators=300, max_depth=15" \
                --comments "Classifier trained on the MNIST dataset, 3rd test"
            # Do something

        """
        options = parse_argstring(self.measure_power, line)
        if options.power_meter not in self.shell.user_ns.keys():
            raise NameError("The PowerMeter variable passed is not referenced")
        self.power_meter = self.shell.user_ns[options.power_meter]
        if not isinstance(self.power_meter, PowerMeter):
            raise TypeError("The first argument is not a PowerMeter object")
        options = vars(options)
        del options["power_meter"]
        with self.power_meter(**options):
            self.shell.run_cell(cell)


def load_ipython_extension(ipython):
    ipython.register_magics(MagicPowerMeter)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Python classes monitoring GPU's power usage
during a time delimited between a start and a stop methods
"""
__all__ = ["NoGpuPower", "NvidiaPower"]

import abc
import logging
import os
import re
import signal
import subprocess
from pathlib import Path

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .utils import TOTAL_ENERGY_GPU, TOTAL_GPU_TIME

LOGGER = logging.getLogger(__name__)

NVIDIAPOWERLOG_FILENAME = "nvidiaPowerLog.csv"


class GpuPower(abc.ABC):
    """
    Class to monitor the power usage of a gpu
    Usage :
    1. initialiase the class
    2. call start
    3. call stop
    4. collect the results with parse_log
    """

    def __init__(self):
        self.record = {TOTAL_GPU_TIME: 0, TOTAL_ENERGY_GPU: 0}

    def start(self):
        """
        Starts the recording processus
        """

    def stop(self):
        """
        Stops the recording processus
        """

    def parse_log(self):
        """
        Extract relevant information from the logs.
        """


class NoGpuPower(GpuPower):
    """
    Class used when no GPU is available
    """

    def __init__(self):
        super().__init__()


class NvidiaPower(GpuPower):
    """
    NVIDIA-based GPU class

    Parameters
    ----------
    interval (int)
        interval to which log power in the log_path file
    """

    def __init__(self, interval=1):
        super().__init__()
        self.log_file = (
            Path(os.path.dirname(os.path.abspath(__file__)))
            / NVIDIAPOWERLOG_FILENAME
        )
        self.logging_process = None
        self.interval = interval

    def start(self):
        """
        Start the measure process
        """
        if self.logging_process is not None:
            self.stop()

        LOGGER.info("starting GPU power monitoring ...")

        self.logging_process = subprocess.Popen(
            [
                "nvidia-smi",
                "-f",
                str(self.log_file),
                "-q",
                "-d",
                "POWER",
                "-l",
                str(self.interval),
            ]
        )

    def stop(self):
        """
        Stop the measure process if started
        """
        LOGGER.info("stopping GPU power monitoring ...")
        self.logging_process.send_signal(signal.SIGINT)
        self.logging_process = None

    def parse_log(self):
        """
        Extract power as logged in the last measure process
        Returns:
        -------
        results (pandas.DataFrame)
            all records with associated ellapsed times
        """
        if not self.log_file.exists():
            raise FileNotFoundError(
                "Logging file not found, make sure you started to \
                    run a measure"
            )
        content = self.log_file.read_text()
        regex_power = r"Power Draw +: (.*) W"
        regex_time = r"Timestamp +: (.*)"
        times = re.findall(regex_time, content)
        powers = re.findall(regex_power, content)
        if len(times) != len(powers):
            min_len = min(len(times), len(powers))
            times = times[:min_len]
            powers = powers[:min_len]
        results = pd.DataFrame(
            np.array([times, powers]).T, columns=["Time", "Power"]
        )
        results["Power"] = results["Power"].astype("float32")
        results["Time"] = pd.to_datetime(
            results["Time"], infer_datetime_format=True
        )
        results["Elapsed time"] = results["Time"] - results.loc[0, "Time"]
        results["Elapsed time"] = results["Elapsed time"].dt.total_seconds()
        self.record[TOTAL_GPU_TIME] = results["Elapsed time"].iloc[-1]
        self.record[TOTAL_ENERGY_GPU] = (
            results["Power"].sum() * self.interval * 1000 / 3600
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['carbonai']

package_data = \
{'': ['*'], 'carbonai': ['data/*']}

install_requires = \
['fuzzywuzzy>=0.17.0',
 'ipython>=7.12',
 'numpy>=1.20.3',
 'pandas>=1.0.5',
 'psutil',
 'requests>=2.20.1']

setup_kwargs = {
    'name': 'carbonai',
    'version': '0.2.1',
    'description': 'Monitor the power consumption of a function',
    'long_description': '# CarbonAI\n\nThis project aims at creating a python package that allows you to monitor the power consumption of any python function.\n\n## Documentation\n\nThe complete documentation is available [here](https://capgemini-invent-france.github.io/CarbonAI/).\n\n## Getting started\n### Install\n\nFirst of all you need to install the intel utility allowing you to monitor power consumption ([support](https://software.intel.com/en-us/articles/intel-power-gadget)):\n* [Windows](https://software.intel.com/file/823776/download)\n* [MacOS](https://software.intel.com/sites/default/files/managed/91/6b/Intel%20Power%20Gadget.dmg)\n* Linux, no need to install any software\n\nTo install this package :\n```sh\npip install carbonai\n```\n\n### Example\n\nThere are several ways to use this package depending on how you develop.\nYou just have to import the `PowerMeter` object, initialize it and call the function you want to monitor.\nPlease insert a description of the running function, the dataset, the model, any info would be useful.\n\n#### Function decorator\nTo monitor the power consumption of a function, follow this example:\n```python\nfrom carbonai import PowerMeter\npower_meter = PowerMeter(project_name="MNIST classifier")\n\n@power_meter.measure_power(\n  package="sklearn",\n  algorithm="RandomForestClassifier",\n  data_type="tabular",\n  data_shape=<your_data>.shape,\n  algorithm_params="n_estimators=300, max_depth=15",\n  comments="Classifier trained on the MNIST dataset, 3rd test"\n)\ndef my_func(arg1, arg2, ...):\n  # Do something\n```\n\n#### Using the with statement\nTo monitor the power consumption of some specific inline code, you can use with statements\n\n```python\nfrom carbonai import PowerMeter\npower_meter = PowerMeter(project_name="MNIST classifier")\n\nwith power_meter(\n  package="sklearn",\n  algorithm="RandomForestClassifier",\n  data_type="tabular",\n  data_shape=<your_data>.shape,\n  algorithm_params="n_estimators=300, max_depth=15",\n  comments="Classifier trained on the MNIST dataset, 3rd test"\n):\n  # Do something\n```\n\n## Contribute\n\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are welcome.\n\nYou can find details on how to contribute in [our guide](CONTRIBUTING.md)\n',
    'author': 'Capgemini Invent - Martin Chauvin, Francois Lemeille, Jordan Toh',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Capgemini-Invent-France/CarbonAI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4',
}


setup(**setup_kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
PowerMeter
---------
Main Python class or entrypoint to monitor the power consumption of
an algorithm.
"""

__all__ = ["PowerMeter"]

import datetime
import getpass
import json
import logging
import os
import shutil
import sys
import traceback
import warnings
from pathlib import Path

import pandas as pd  # type: ignore
import requests

from .nvidia_power import NoGpuPower, NvidiaPower
from .power_gadget import (
    NoPowerGadget,
    PowerGadgetLinuxMSR,
    PowerGadgetLinuxRAPL,
    PowerGadgetMac,
    PowerGadgetWin,
)
from .utils import (
    AVAILABLE_STEPS,
    COUNTRY_CODE_COLUMN,
    COUNTRY_NAME_COLUMN,
    ENERGY_MIX_COLUMN,
    ENERGY_MIX_DATABASE,
    LOGGING_FILE,
    MSR_PATH_LINUX_TEST,
    PACKAGE_PATH,
    POWERLOG_PATH_LINUX,
    TOTAL_CPU_TIME,
    TOTAL_ENERGY_ALL,
    TOTAL_ENERGY_CPU,
    TOTAL_ENERGY_GPU,
    TOTAL_ENERGY_MEMORY,
    TOTAL_ENERGY_PROCESS_CPU,
    TOTAL_ENERGY_PROCESS_MEMORY,
    TOTAL_GPU_TIME,
    match,
    normalize,
)

LOGGER = logging.getLogger(__name__)


class PowerMeter:
    """
    PowerMeter is a general tool to monitor and log the power consumption of
    any given function.

    Depending on the platform and hardware used, it allows to measure the
    power usage of the CPU,the DRAM and the GPU. The measure is, then,
    converted to CO2 emissions depending on the country set and the PUE
    of the machine (background on the PUE `here`_). You can chose to log the
    results locally or send it to an endpoint.

    Parameters
    ----------
    project_name : str, default current_directory_name
        Name of the project you are working on.
    program_name : str, optional
        Name of the program you are working on.
    client_name : str, optional
        Name of the client you are working for.
    user_name : str, default computer_user_name
        The name of the user using the tool (for logging purpose).
    cpu_power_log_path : str, optional
        The path to the tool "PowerLog" or "powercap" or "IntelPowerGadget".
    get_country : bool, default True
        Whether to retrieve user country location or not (uses the user IP).
    location : str, optional
        Country ISO Code available `here
        <https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements>`_

        Must be set if get_country is False.
    is_online : boolean, default True
        Whether the machine is connected to the internet or not.
    filepath : str, default .
        Path of the file where all the carbon logs will be written.
    output_format : {'csv', 'excel'}, default 'csv'
        Format of the carbon logs file produced.
    api_endpoint : str, optional
        Endpoint of the API to upload the collected data to

        Note: we provide an endpoint to collect data and contribute, with the
        community, towards greener algorithms.
        Here is the url :
        https://ngji0jx9dc.execute-api.eu-west-3.amazonaws.com/post_new_item

    See Also
    --------
    PowerMeter.from_config : Create a power meter from a config file.

    Examples
    --------
    Create a PowerMeter for the project `MNIST Classifier` while
    not being online.

    >>> power_meter = PowerMeter(project_name="MNIST classifier",
    ...     is_online=False, location="FR")

    Create a PowerMeter for the project `Test` and send the collected data to
    our endpoint.

    >>> power_meter = PowerMeter(project_name="Test",
    ...     api_endpoint="https://ngji0jx9dc.execute-
    api.eu-west-3.amazonaws.com/post_new_item")

    Notes
    -----
    This package may log private data (username, country, project_name).
    If you do not provide any api_endpoint, we will never have access
    to this data.

    On the other hand, if you chose to share your data with us
    (by using our endpoint:
    https://ngji0jx9dc.execute-api.eu-west-3.amazonaws.com/post_new_item
    ), we commit to anonymize any data shared.
    """

    LAPTOP_PUE = 1.3  # pue for my laptop
    SERVER_PUE = 1.58  # pue for a server
    DEFAULT_LOCATION = "FR"
    DATETIME_FORMAT = "%m/%d/%Y %H:%M:%S"  # "%c"

    # ----------------------------------------------------------------------
    # Constructors
    def __init__(
        self,
        project_name="",
        program_name="",
        client_name="",
        user_name="",
        cpu_power_log_path="",
        get_country=True,
        location="",
        is_online=True,
        filepath=None,
        output_format="csv",
        api_endpoint=None,
    ):

        self.platform = sys.platform
        powergadget_platform = {
            "darwin": PowerGadgetMac,
            "win32": PowerGadgetWin,
            "linux": self.__set_powergadget_linux,
            "linux2": self.__set_powergadget_linux,
            "": NoPowerGadget,
        }
        self.power_gadget = powergadget_platform[self.platform](
            powerlog_path=cpu_power_log_path
        )

        self.pue = self.__set_pue()

        self.cuda_available = self.__check_gpu()
        self.gpu_power = self.__set_gpu_power()

        self.user = self.__set_username(user_name)

        self.project = self.__set_project_name(project_name)
        self.program_name = self.__set_project_entity(program_name)
        self.client_name = self.__set_project_entity(client_name)

        self.is_online = is_online
        if api_endpoint:
            LOGGER.info("Api endpoint given, will save data online")
            self.api_endpoint = api_endpoint
        else:
            LOGGER.info("No current api endpoint, will save data locally")
            self.api_endpoint = ""

        self.location = self.__set_location(location, get_country)

        self.energy_mix_db = self.__load_energy_mix_db()
        self.energy_mix = self.__get_energy_mix()  # kgCO2e/kWh
        self.location_name = self.__get_location_name()

        self.used_package = ""
        self.used_algorithm = ""
        self.used_data_type = ""
        self.used_data_shape = ""
        self.used_algorithm_params = ""
        self.used_comments = ""
        self.used_step = ""

        if not filepath:
            LOGGER.info("No current filepath, will use the default")
            self.filepath = Path.cwd() / "emissions.csv"
        else:
            LOGGER.info("Filepath given, will save there")
            self.filepath = Path(filepath)

        self.output_format = output_format

        self.logging_filename = PACKAGE_PATH / LOGGING_FILE

    @staticmethod
    def __load_energy_mix_db():
        return pd.read_csv(
            PACKAGE_PATH / ENERGY_MIX_DATABASE, encoding="utf-8"
        )

    @staticmethod
    def __extract_env_name():
        env = "unknown"
        try:
            env = os.environ["CONDA_DEFAULT_ENV"]
        except KeyError:
            pass
        if hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix:
            env = sys.base_prefix.split("/")[-1]
        return env

    @staticmethod
    def __check_gpu():
        cuda_available = False
        if shutil.which("nvidia-smi"):
            cuda_available = True

        return cuda_available

    @staticmethod
    def __get_country():
        """
        Retrieve the ISO code country
        Beware of the encoding
        cf. from https://stackoverflow.com/questions/40059654/python-convert-
        a-bytes-array-into-json-format
        """
        request = requests.get("http://ipinfo.io/json")
        response = request.content.decode("utf8").replace("'", '"')
        user_info = json.loads(response)
        return user_info["country"]

    @classmethod
    def from_config(cls, path):
        """
        Create a PowerMeter from a json config file.

        The json config file may contain any key named after the arguments of
        the :class:`carbonai.PowerMeter` constructor.

        Parameters
        ----------
        path : str
            Path to the json config file

        Returns
        -------
        PowerMeter

        See also
        --------
        PowerMeter : Instantiate a PowerMeter by declaring every argument

        Examples
        --------
        Example of config file used. This file is named `config.json`

        .. code-block:: JSON

            {
                "project_name": "Project X",
                "program_name": "Program X",
                "client_name": "Client X",
                "get_country": true,
                "is_online": false,
                "user_name": "customUsername",
                "filepath": "",
                "api_endpoint": "...",
                "location":"FR"
            }

        To load this file as a power meter.

        >>> power_meter = PowerMeter.from_config("config.json")
        """
        with open(path) as file:
            args = json.load(file)
        return cls(**args)

    def __set_pue(self):
        if self.platform in ["darwin", "win32"]:
            return self.LAPTOP_PUE  # pue for my laptop
        return self.SERVER_PUE  # pue for a server

    def __set_powergadget_linux(self, powerlog_path=None):
        if POWERLOG_PATH_LINUX.exists():
            power_gadget = PowerGadgetLinuxRAPL()
        # The user needs to be root to use MSR interface
        elif MSR_PATH_LINUX_TEST.exists() and os.getuid() == 0:
            power_gadget = PowerGadgetLinuxMSR()
        else:
            LOGGER.warning("No power reading interface was found")
            power_gadget = NoPowerGadget()
        return power_gadget

    def __set_gpu_power(self):
        if self.cuda_available:
            LOGGER.info("Found a GPU")
            gpu_power = NvidiaPower()
        else:
            LOGGER.info("Found no GPU")
            gpu_power = NoGpuPower()
        return gpu_power

    def __set_username(self, user_name):
        if user_name:
            return user_name
        return getpass.getuser()

    def __set_project_name(self, project_name):
        if project_name:
            return project_name
        return self.__extract_env_name()

    def __set_project_entity(self, entity_name):
        if entity_name:
            return entity_name
        return "--"

    def __set_location(self, provided_location, get_country):
        # Set the location used to convert energy usage to carbon emissions
        # if the location is provided, we use it
        # if it's not and we can use the internet and the user authorize us to
        # #do so then we retrieve it from the IP address
        # otherwise set the location to default
        if provided_location:
            location = provided_location
        elif (self.is_online or self.api_endpoint) and get_country:
            location = self.__get_country()
        else:
            warnings.warn(
                "No location was set, we will fallback to \
                the default location: {}".format(
                    self.DEFAULT_LOCATION
                )
            )
            location = self.DEFAULT_LOCATION
        return location

    def __get_energy_mix(self):
        if not (
            self.energy_mix_db[COUNTRY_CODE_COLUMN] == self.location
        ).any():
            raise NameError(
                "The location input was not found, make sure you wrote "
                "the isocode of your country. You used " + self.location
            )
        return self.energy_mix_db.loc[
            self.energy_mix_db[COUNTRY_CODE_COLUMN] == self.location,
            ENERGY_MIX_COLUMN,
        ].values[0]

    def __get_location_name(self):
        return self.energy_mix_db.loc[
            self.energy_mix_db[COUNTRY_CODE_COLUMN] == self.location,
            COUNTRY_NAME_COLUMN,
        ].values[0]

    def __aggregate_power(self, cpu_record, gpu_record):
        """
        Implement the cO2 emission value

        Parameters:
        -----------
        cpu_record, gpu_record (dict):
            Respectively CPU and GPU's records

        Returns
        -------
        co2_emitted (float)
        """
        used_energy = self.pue * (
            cpu_record[TOTAL_ENERGY_PROCESS_CPU]
            + cpu_record[TOTAL_ENERGY_PROCESS_MEMORY]
            + gpu_record[TOTAL_ENERGY_GPU]
        )  # mWh
        co2_emitted = used_energy * self.energy_mix * 1e-3
        LOGGER.info(
            "This process emitted %.3fg of CO2 (using the energy mix of %s)",
            co2_emitted,
            self.location_name,
        )

        return co2_emitted

    def measure_power(
        self,
        package,
        algorithm,
        step="other",
        data_type="",
        data_shape="",
        algorithm_params="",
        comments="",
    ):
        """
        A decorator to measure the power consumption of a given function

        Parameters
        ----------
        package : str
            A string describing the package used by this function
            (e.g. sklearn, Pytorch, ...)
        algorithm : str
            A string describing the algorithm used in the function monitored
            (e.g. RandomForestClassifier, ResNet121, ...)
        step : {'inference', 'training', 'other', 'test', 'run', \
            'preprocessing'}, optional
            A string to provide useful information on the current stage
            of the algorithm
        data_type : {'tabular', 'image', 'text', 'time series', 'other'},\
             optional
            A string describing the type of data used for training
        data_shape : str or tuple, optional
            A string or tuple describing the quantity of data used
        algorithm_params : str, optional
            A string describing the parameters used by the algorithm
        comments : str, optional
            A string to provide any useful information

        Returns
        -------

        See also
        --------
        PowerMeter.from_config : Create a PowerMeter object from a config file
        PowerMeter.__call__ : Measure the power usage using a with statement

        Examples
        --------
        First, create a PowerMeter (you only do to this step once).

        >>> power_meter = PowerMeter.from_config("config.json")

        Decorate the function you wish to monitor.

        >>> @power_meter.measure_power(
        ...     package="pandas, numpy",
        ...     algorithm="data cleaning",
        ...     step="preprocessing",
        ...     data_type="tabular",
        ...     comments="Cleaning of csv files + train-test splitting"
        ... )
        ... def example_func():
        ...     # do something

        Each time this function will be called, the PowerMeter will monitor
        the power usage and log the function's carbon footprint.

        >>> example_func()
        result_of_your_function
        """
        if not algorithm or not package:
            raise SyntaxError(
                "Please input a description for the function you are trying "
                "to monitor. Pass in the algorithm and the package you are "
                "trying to monitor"
            )

        def decorator(func):
            def wrapper(*args, **kwargs):
                self.start_measure(
                    package,
                    algorithm,
                    data_type=data_type,
                    data_shape=data_shape,
                    algorithm_params=algorithm_params,
                    comments=comments,
                    step=step,
                )
                try:
                    results = func(*args, **kwargs)
                finally:
                    self.stop_measure()
                return results

            return wrapper

        return decorator

    def __set_used_arguments(
        self,
        package,
        algorithm,
        data_type="",
        data_shape="",
        algorithm_params="",
        comments="",
        step="other",
    ):
        self.used_package = normalize(package)
        self.used_algorithm = normalize(algorithm)
        self.used_data_type = normalize(data_type)
        self.used_data_shape = normalize(data_shape)
        self.used_algorithm_params = normalize(algorithm_params)
        self.used_comments = normalize(comments)
        self.used_step = normalize(match(step, AVAILABLE_STEPS))

    def __call__(
        self,
        package,
        algorithm,
        step="other",
        data_type="",
        data_shape="",
        algorithm_params="",
        comments="",
    ):
        """
        Measure the power usage using a with statement.

        Parameters
        ----------
        package : str
            A string describing the package used by this function (e.g.
            sklearn, Pytorch, ...)
        algorithm : str
            A string describing the algorithm used in the function monitored
            (e.g. RandomForestClassifier, ResNet121, ...)
        step : {'inference', 'training', 'other', 'test', 'run', \
            'preprocessing'}, optional
            A string to provide useful information on the current stage
            of the alogrithm
        data_type : {'tabular', 'image', 'text', 'time series', 'other'}, \
            optional
            A string describing the type of data used for training
        data_shape : str or tuple, optional
            A string or tuple describing the quantity of data used
        algorithm_params : str, optional
            A string describing the parameters used by the algorithm
        comments : str, optional
            A string to provide any useful information

        Returns
        -------

        See also
        --------
        PowerMeter.from_config : Create a PowerMeter object from a config file
        PowerMeter.measure_power : Measure the power usage using a function \
            decorator

        Examples
        --------
        First, create a PowerMeter (you only do to this step once).

        >>> power_meter = PowerMeter.from_config("config.json")

        Use a with statement to encapsulate the code you want to monitor

        >>> with power_meter(
        ...     package="pandas, numpy",
        ...     algorithm="data cleaning",
        ...     step="preprocessing",
        ...     data_type="tabular",
        ...     comments="Cleaning of csv files + train-test splitting"
        ... ):
        ...     # do something
        result_of_your_code
        """
        self.__set_used_arguments(
            package,
            algorithm,
            data_type=data_type,
            data_shape=data_shape,
            algorithm_params=algorithm_params,
            comments=comments,
            step=step,
        )
        return self

    def __enter__(
        self,
    ):
        self.start_measure(
            self.used_package,
            self.used_algorithm,
            data_type=self.used_data_type,
            data_shape=self.used_data_shape,
            algorithm_params=self.used_algorithm_params,
            comments=self.used_comments,
        )

    def __exit__(self, exit_type, value, traceback):
        self.stop_measure()

    def start_measure(
        self,
        package,
        algorithm,
        step="other",
        data_type="",
        data_shape="",
        algorithm_params="",
        comments="",
    ):
        """
        Starts mesuring the power consumption of a given sample of code

        Parameters
        ----------
        package : str
            A string describing the package used by this function
            (e.g. sklearn, Pytorch, ...)
        algorithm : str
            A string describing the algorithm used in the function monitored
            (e.g. RandomForestClassifier, ResNet121, ...)
        step : {'inference', 'training', 'other', 'test', 'run', \
            'preprocessing'}, optional
            A string to provide useful information on the current stage
            of the algorithm
        data_type : {'tabular', 'image', 'text', 'time series', 'other'},\
            optional
            A string describing the type of data used for training
        data_shape : str or tuple, optional
            A string or tuple describing the quantity of data used
        algorithm_params : str, optional
            A string describing the parameters used by the algorithm
        comments : str, optional
            A string to provide any useful information

        Returns
        -------

        See also
        --------
        PowerMeter.stop_measure : Stop the measure started with start_measure
        PowerMeter.from_config : Create a PowerMeter object from a config file
        PowerMeter.measure_power : Measure the power usage using a function \
            decorator
        PowerMeter.__call__ : Measure the power usage using a with statement

        Notes
        -----
        We do not recommend using this method to monitor the energy usage of
        your code because it won't automatically stop if an error is raised
        at some point while running. You will then have to stop the measure
        manually with :func:`PowerMeter.stop_measure`.

        Examples
        --------
        First, create a PowerMeter (you only do to this step once).

        >>> power_meter = PowerMeter.from_config("config.json")

        Start measuring the code you wish to monitor

        >>> power_meter.start_measure(
        ...     package="pandas, numpy",
        ...     algorithm="data cleaning",
        ...     step="preprocessing",
        ...     data_type="tabular",
        ...     comments="Cleaning of csv files + train-test splitting"
        ... )
        ... # do something
        result_of_your_code

        **Do not forget to stop measuring**

        >>> power_meter.stop_measure()

        """
        self.gpu_power.start()
        self.power_gadget.start()
        self.__set_used_arguments(
            package,
            algorithm,
            data_type=data_type,
            data_shape=data_shape,
            algorithm_params=algorithm_params,
            comments=comments,
            step=step,
        )

    def stop_measure(self):
        """
        Stops the measure started with :func:`PowerMeter.start_measure`

        Parameters
        ----------

        See also
        --------
        PowerMeter.start_measure : Stop the measure started with start_measure
        PowerMeter.from_config : Create a PowerMeter object from a config file
        PowerMeter.measure_power : Measure the power usage using a function \
            decorator
        PowerMeter.__call__ : Measure the power usage using a with statement

        Notes
        -----
        We do not recommend using this method to monitor the energy usage of
        your code because it won't automatically stop if an error is raised at
        some point while running. You will then have to stop the measure
        manually with :func:`PowerMeter.stop_measure`.

        Examples
        --------
        First, create a PowerMeter (you only do to this step once).

        >>> power_meter = PowerMeter.from_config("config.json")

        Start measuring the code you wish to monitor

        >>> power_meter.start_measure(
        ...     package="pandas, numpy",
        ...     algorithm="data cleaning",
        ...     step="preprocessing",
        ...     data_type="tabular",
        ...     comments="Cleaning of csv files + train-test splitting"
        ... )
        ... # do something
        result_of_your_code

        **Do not forget to stop measuring**

        >>> power_meter.stop_measure()
        """
        self.power_gadget.stop()
        self.gpu_power.stop()
        self.gpu_power.parse_log()
        self.__log_records(
            self.power_gadget.record,  # must be a dict
            self.gpu_power.record,  # must be a dict
            algorithm=self.used_algorithm,
            package=self.used_package,
            data_type=self.used_data_type,
            data_shape=self.used_data_shape,
            algorithm_params=self.used_algorithm_params,
            comments=self.used_comments,
            step=self.used_step,
        )

    def __record_data_to_server(self, info):
        headers = {"Content-Type": "application/json"}
        data = json.dumps(info)
        try:
            response = requests.request(
                "POST",
                self.api_endpoint,
                headers=headers,
                data=data,
                timeout=1,
            )
            return response.status_code
        except requests.exceptions.Timeout:
            return 408

    def __record_data_to_csv_file(self, info):
        try:
            data = pd.DataFrame(info, index=[0])
            if Path(self.filepath).exists():
                data.to_csv(self.filepath, mode="a", index=False, header=False)
            else:
                data.to_csv(self.filepath, index=False)
            return True
        except Exception:
            LOGGER.error("* error during the csv writing process *")
            LOGGER.error(traceback.format_exc())
            return False

    def __record_data_to_excel_file(self, info):
        try:

            if Path(self.filepath).exists():
                data = pd.read_excel(self.filepath).append(
                    info, ignore_index=True
                )
                data.to_excel(self.filepath, index=False)
            else:
                data = pd.DataFrame(info, index=[0])
                data.to_excel(self.filepath, index=False)
            return True
        except Exception:
            LOGGER.error(
                "* error during the writing process in an excel file *"
            )
            LOGGER.error(traceback.format_exc())
            return False

    def __record_data_to_file(self, info):
        """
        Only two options so far: CSV or EXCEL
        """
        if self.filepath.suffix == ".csv":
            return self.__record_data_to_csv_file(info)
        elif self.filepath.suffix == ".xls" or self.filepath.suffix == ".xlsx":
            return self.__record_data_to_excel_file(info)
        LOGGER.info("unknown format: it should be either .csv, .xls or .xlsx")
        return self.__record_data_to_excel_file(info)

    def __log_records(
        self,
        cpu_recorded_power,
        gpu_recorded_power,
        algorithm="",
        package="",
        data_type="",
        data_shape="",
        algorithm_params="",
        comments="",
        step="other",
    ):
        co2_emitted = self.__aggregate_power(
            self.power_gadget.record, self.gpu_power.record
        )
        payload = {
            "Datetime": datetime.datetime.now().strftime(self.DATETIME_FORMAT),
            "Country": self.location_name,
            "Platform": self.platform,
            "User ID": self.user,
            "ISO": self.location,
            "Project name": self.project,
            "Program name": self.program_name,
            "Client name": self.client_name,
            "Total Elapsed CPU Time (sec)": cpu_recorded_power[TOTAL_CPU_TIME],
            "Total Elapsed GPU Time (sec)": gpu_recorded_power[TOTAL_GPU_TIME],
            "Cumulative Package Energy (mWh)": cpu_recorded_power[
                TOTAL_ENERGY_ALL
            ],
            "Cumulative IA Energy (mWh)": cpu_recorded_power[TOTAL_ENERGY_CPU],
            "Cumulative GPU Energy (mWh)": gpu_recorded_power[
                TOTAL_ENERGY_GPU
            ],
            "Cumulative DRAM Energy (mWh)": cpu_recorded_power[
                TOTAL_ENERGY_MEMORY
            ],
            "Cumulative process CPU Energy (mWh)": cpu_recorded_power[
                TOTAL_ENERGY_PROCESS_CPU
            ],
            "Cumulative process DRAM Energy (mWh)": cpu_recorded_power[
                TOTAL_ENERGY_PROCESS_MEMORY
            ],
            "PUE": self.pue,
            "CO2 emitted (gCO2e)": co2_emitted,
            "Package": package,
            "Algorithm": algorithm,
            "Algorithm's parameters": algorithm_params,
            "Data type": data_type,
            "Data shape": data_shape,
            "Comment": comments,
            "Step": step,
        }
        written = self.__record_data_to_file(payload)
        LOGGER.info("* recorded into a file? %s*", written)

        if self.is_online and self.api_endpoint:
            response_status_code = self.__record_data_to_server(payload)
            if response_status_code >= 400:
                LOGGER.warning(
                    "We couldn't upload the recorded data to the server, we"
                    "are going to record it for a later upload"
                )
                # can't upload we'll record the data
                data = pd.DataFrame(payload, index=[0])
                if self.logging_filename.exists():
                    data.to_csv(
                        self.logging_filename,
                        mode="a",
                        index=False,
                        header=False,
                    )
                else:
                    data.to_csv(self.logging_filename, index=False)

            if response_status_code == 200:
                # we successfully uploaded,
                # check if there are other locally recorded data
                if self.logging_filename.exists():
                    data = pd.read_csv(self.logging_filename, index_col=None)
                    payloads = data.to_dict(orient="records")
                    for i, payload in enumerate(payloads):
                        res_status_code = self.__record_data_to_server(payload)
                        if res_status_code != 200:
                            data.iloc[i:].to_csv(
                                self.logging_filename, index=False
                            )
                            break
                    else:
                        self.logging_filename.unlink()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Python wrapper of Intel(R) Power Gadget, a software-based power estimation tool
enabled for 2nd Generation Intel(R) Core(TM) processors or later.
The software data can be logged from the command line (without running the GUI
app) using PowerLog.
"""
__all__ = [
    "PowerGadgetMac",
    "PowerGadgetWin",
    "PowerGadgetLinuxRAPL",
    "PowerGadgetLinuxMSR",
    "NoPowerGadget",
]

import abc
import datetime
import glob
import logging
import os
import re
import struct
import subprocess
import threading
import time
from pathlib import Path

import pandas as pd  # type: ignore
import psutil  # type: ignore

from .utils import (
    HOME_DIR,
    MAC_INTELPOWERLOG_FILENAME,
    PACKAGE_PATH,
    TOTAL_CPU_TIME,
    TOTAL_ENERGY_ALL,
    TOTAL_ENERGY_CPU,
    TOTAL_ENERGY_MEMORY,
    TOTAL_ENERGY_PROCESS_CPU,
    TOTAL_ENERGY_PROCESS_MEMORY,
    WIN_INTELPOWERLOG_FILENAME,
)

LOGGER = logging.getLogger(__name__)

DMG_PATH = Path("/tmp/IntelPowerGadget.dmg")
MSI_PATH = Path("C:\\tmp\\IntelPowerGadget.msi")

POWERLOG_PATH_MAC = Path("/Applications/Intel Power Gadget/PowerLog")
POWERLOG_PATH_WIN = [
    Path("/Program Files/Intel/Power Gadget 3.6"),
    Path("/Program Files/Intel/Power Gadget 3.5"),
]
POWERLOG_TOOL_WIN = "IntelPowerGadget.exe"


CPU_IDS_DIR = "/sys/devices/system/cpu/cpu*/topology/physical_package_id"
READ_MSR_PATH = "/dev/cpu/{}/msr"
READ_RAPL_PATH = (
    "/sys/class/powercap/intel-rapl/intel-rapl:{}/"  # rapl_socket_id
)
RAPL_DEVICENAME_FILE = "name"
RAPL_ENERGY_FILE = "energy_uj"
RAPL_DRAM_PATH = "intel-rapl:{}:{}/"  # rapl_socket_id, rapl_device_id


class PowerGadget(abc.ABC):
    """
    Abstract wrapper to an Intel Power Gadget object instanciated \
        by a command line interface
    """

    @staticmethod
    def parse_log(powerlog_file, process_usage=None):
        """
        From the file made by PowerLog, we extract relevant information.

        Parameters
        ----------
        powerlog_file
            Pathlib.Path instance
        Returns
        -------
        results (dict)
            relevant information such as "Cumulative Package Energy (mWh)",
            "Cumulative IA Energy (mWh)" or "Cumulative GPU Energy (mWh)"
        """
        content = powerlog_file.read_text()
        if "Total Elapsed Time" not in content:
            LOGGER.debug(
                "The log file does not seem to be written yet, \
                we'll wait 2 secs."
            )
            time.sleep(2)
            content = powerlog_file.read_text()
        results = {
            TOTAL_CPU_TIME: 0,
            TOTAL_ENERGY_ALL: 0,
            TOTAL_ENERGY_CPU: 0,
            TOTAL_ENERGY_MEMORY: 0,
        }
        results[TOTAL_CPU_TIME] = float(
            re.search(
                r"(?<=Total Elapsed Time \(sec\) = )(\d|\.)*", content
            ).group(0)
        )
        results[TOTAL_ENERGY_ALL] = float(
            re.search(
                r"((?<=Cumulative Package Energy_0 \(mWh\) = )|"
                r"(?<=Cumulative Processor Energy_0 \(mWh\) = ))(\d|\.)*",
                content,
            ).group(0)
        )
        results[TOTAL_ENERGY_CPU] = float(
            re.search(
                r"(?<=Cumulative IA Energy_0 \(mWh\) = )(\d|\.)*", content
            ).group(0)
        )
        results[TOTAL_ENERGY_MEMORY] = float(
            re.search(
                r"(?<=Cumulative DRAM Energy_0 \(mWh\) = )(\d|\.)*", content
            ).group(0)
        )
        if process_usage:
            # to account for the actual algorithm energy consumption we need to
            # combine the overall mesure of the machine power usage with
            # the actual algorithm CPU and memory usage
            process_usage = pd.DataFrame(
                process_usage,
                columns=["time", "process_cpu_usage", "process_memory_usage"],
            )
            powers = pd.read_csv(powerlog_file)
            powers = powers.dropna(
                subset=["Cumulative Processor Energy_0(mWh)"]
            )
            powers["System Time"] = pd.to_datetime(
                powers["System Time"], format="%H:%M:%S:%f"
            )
            powers_sec = powers.groupby(
                pd.Grouper(key="System Time", freq="S")
            )[
                [
                    "Cumulative Processor Energy_0(mWh)",
                    "Cumulative IA Energy_0(mWh)",
                    "Cumulative DRAM Energy_0(mWh)",
                ]
            ].apply(
                lambda x: x.iloc[-1] - x.iloc[0]
            )
            process_usage["time"] = process_usage["time"].dt.floor("S").dt.time
            process_usage.set_index("time", inplace=True)
            powers_sec.index = powers_sec.index.time
            power_process = pd.merge(
                powers_sec,
                process_usage,
                how="left",
                left_index=True,
                right_index=True,
            )
            # a measure is performed each second but
            # it actually takes a little more than 1s
            # so when merging on the timestamp there may be some empty values
            #  that we fill with the previous one
            power_process[
                ["process_cpu_usage", "process_memory_usage"]
            ] = power_process[
                ["process_cpu_usage", "process_memory_usage"]
            ].fillna(
                method="ffill", limit=1
            )
            results[TOTAL_ENERGY_PROCESS_CPU] = (
                power_process["Cumulative IA Energy_0(mWh)"]
                * power_process["process_cpu_usage"]
            ).sum()
            results[TOTAL_ENERGY_PROCESS_MEMORY] = (
                power_process["Cumulative DRAM Energy_0(mWh)"]
                * power_process["process_memory_usage"]
            ).sum()
        return results

    def __init__(self):
        self.record = {}
        self.thread = None

    def __get_powerlog_file(self):
        """
        Retrieve the log file where is written the PowerLog logs
        """

    def start(self):
        """
        Starts the recording processus with Intel Power Gadget
        """

    def stop_thread(self):
        self.thread.do_run = False
        self.thread.join()

    def get_computer_usage(self, process, interval=1):
        """Compute the ratio of cpu and memory used by the current process

        Parameters
        ----------
        process : psutil.Process
            the current process
        interval : int
            interval at which measure ratios

        Returns
        -------
        Tuple
            time of execution, ratio of cpu used, ration of memory used
        """
        with process.oneshot():
            process_cpu_usage = process.cpu_percent(interval=interval)
            cpu_usage = psutil.cpu_percent()
            process_cpu_usage = process_cpu_usage / (
                cpu_usage * psutil.cpu_count()
            )
            memory_global = psutil.virtual_memory()
            memory_usage = process.memory_full_info().rss / (
                memory_global.total - memory_global.available
            )
        return datetime.datetime.now(), process_cpu_usage, memory_usage

    def stop(self):
        """
        Stops the recording processus with Intel Power Gadget
        """


class NoPowerGadget(PowerGadget):
    """
    Dummy class used when no power gadget (intel power gadget, rapl, msr)
    is available. It will return empty consumption
    """

    def __init__(self):
        super().__init__()
        self.start_time = 0

    def start(self):
        self.start_time = time.time()

    def stop(self):
        end_time = time.time()
        self.record[TOTAL_ENERGY_CPU] = 0
        self.record[TOTAL_ENERGY_PROCESS_CPU] = 0
        self.record[TOTAL_ENERGY_PROCESS_MEMORY] = 0
        self.record[TOTAL_ENERGY_MEMORY] = 0
        self.record[TOTAL_CPU_TIME] = end_time - self.start_time
        self.record[TOTAL_ENERGY_ALL] = 0


class PowerGadgetMac(PowerGadget):
    """
    Mac OS X custom PowerGadget wrapper.
    """

    def __init__(self, powerlog_path=""):
        super().__init__()
        if powerlog_path:
            self.powerlog_path = Path(powerlog_path)
        else:
            self.powerlog_path = POWERLOG_PATH_MAC

        if not self.powerlog_path.exists():
            raise ModuleNotFoundError(
                "We didn't find the Intel Power Gadget tool.\n"
                "Make sure it is installed (download available here : "
                "https://software.intel.com/sites/default/files/managed/91/6b/"
                "Intel%20Power%20Gadget.dmg).\n"
                "If it is installed, we looked for it here: "
                f"{powerlog_path}, try passing the path to "
                "the powerLog tool to the powerMeter."
            )
        self.powerlog_file = self.__get_powerlog_file()
        # self.thread = None
        self.power_draws = []

    def __get_powerlog_file(self):
        return PACKAGE_PATH / MAC_INTELPOWERLOG_FILENAME

    def __get_power_consumption(self, duration=1, resolution=500):
        """
        Retrieve all the available power consumptions with PowerLog using the
        following cli: PowerLog [-resolution ] -duration  [-verbose] [-file ]

        Parameters:
        -----------
        duration, resolution (int):
            cli's arguments

        Returns:
        --------
        consumption (dict)
        """
        _ = subprocess.run(
            [
                self.powerlog_path,
                "-resolution",
                str(resolution),
                "-duration",
                str(duration),
                "-file",
                self.powerlog_file,
            ],
            stdout=open(os.devnull, "wb"),
            check=True,
        )
        consumption = self.parse_log(self.powerlog_file)
        return consumption

    def __append_energy_usage(self, process, interval=1):
        energy_usage = self.__get_power_consumption(duration=interval)
        _, cpu_usage, memory_usage = self.get_computer_usage(
            process, interval=0
        )
        energy_usage[TOTAL_ENERGY_PROCESS_CPU] = (
            energy_usage[TOTAL_ENERGY_CPU] * cpu_usage
        )
        energy_usage[TOTAL_ENERGY_PROCESS_MEMORY] = (
            energy_usage[TOTAL_ENERGY_MEMORY] * memory_usage
        )
        self.power_draws.append(energy_usage)

    def get_power_consumption(self, interval=1):
        """

        Parameters
        ----------
        interval (int)
        """
        current_process = psutil.Process()
        _, _, _ = self.get_computer_usage(
            current_process, interval=0
        )  # initialize the cpu usage
        self.__append_energy_usage(current_process, interval=interval)
        while getattr(self.thread, "do_run", True):
            self.__append_energy_usage(current_process, interval=interval)
        self.__append_energy_usage(current_process, interval=interval)

    def start(self):
        LOGGER.info("starting CPU power monitoring ...")
        if self.thread and self.thread.is_alive():
            self.stop_thread()
        self.power_draws = []
        self.thread = threading.Thread(
            target=self.get_power_consumption, args=()
        )
        self.thread.start()

    def stop(self):
        LOGGER.info("stoping CPU power monitoring ...")
        self.stop_thread()
        self.record = pd.DataFrame.from_records(self.power_draws)
        self.record = self.record.sum(axis=0)


class PowerGadgetWin(PowerGadget):
    """
    Windows custom PowerGadget wrapper.
    """

    def __init__(self, powerlog_path=""):
        super().__init__()
        if powerlog_path:
            self.powerlog_path = Path(powerlog_path)
        else:
            # Try different version of the powerlog path on windows
            for path_win in POWERLOG_PATH_WIN:
                if path_win.exists():
                    self.powerlog_path = path_win / POWERLOG_TOOL_WIN
                    break
        if not self.powerlog_path.exists():
            raise ModuleNotFoundError(
                "We didn't find the Intel Power Gadget tool.\n"
                "Make sure it is installed (download available here :"
                "https://software.intel.com/sites/default/files/managed/91/6b/"
                "Intel%20Power%20Gadget.dmg).\n"
                "If it is installed, we looked for it here:"
                f"{self.powerlog_path}, try passing the path to the "
                "powerLog tool to the powerMeter."
            )
        self.thread = None
        self.process_usage = []

    def __get_powerlog_file(self):
        file_names = glob.glob(
            str(HOME_DIR / "Documents" / WIN_INTELPOWERLOG_FILENAME)
        )
        file_names.sort(
            key=lambda f: list(map(int, re.split(r"-|_|\.|/", f)[-7:-1]))
        )
        return Path(file_names[-1])

    def get_process_usage(self, interval=1):
        current_process = psutil.Process()
        # called once to initialize the cpu monitoring
        psutil.cpu_percent()
        while getattr(self.thread, "do_run", True):
            self.process_usage.append(
                self.get_computer_usage(current_process, interval=interval)
            )
        else:
            self.process_usage.append(
                self.get_computer_usage(current_process, interval=interval)
            )

    def start(self):
        LOGGER.info("starting CPU power monitoring ...")
        if self.thread and self.thread.is_alive():
            LOGGER.debug(
                "another thread is alive, we are going to close it first"
            )
            self.stop_thread()
        _ = subprocess.Popen(
            '"' + str(self.powerlog_path) + '" /min',
            stdin=None,
            stdout=None,
            stderr=None,
            shell=True,
        )
        time.sleep(1)
        self.thread = threading.Thread(target=self.get_process_usage, args=())
        self.thread.start()
        _ = subprocess.run(
            '"' + str(self.powerlog_path) + '" -start', shell=True, check=True
        )

    def stop(self):
        LOGGER.info("stoping CPU power monitoring ...")
        _ = subprocess.run(
            '"' + str(self.powerlog_path) + '" -stop', shell=True, check=True
        )
        _ = subprocess.run(
            'taskkill /IM "' + POWERLOG_TOOL_WIN + '"',
            stdout=open(os.devnull, "wb"),
            shell=True,
            check=True,
        )
        self.stop_thread()
        powerlog_file = self.__get_powerlog_file()
        self.record = self.parse_log(
            powerlog_file, process_usage=self.process_usage
        )
        os.remove(powerlog_file)


class PowerGadgetLinux(PowerGadget):
    """
    Linux custom PowerGadget wrapper.
    """

    @staticmethod
    def __get_cpu_ids():
        """
        Returns the cpu id of this machine
        """
        cpu_ids = []
        for filename in glob.glob(CPU_IDS_DIR):
            with open(filename, "r") as f:
                package_id = int(f.read())
            if package_id not in cpu_ids:
                cpu_ids.append(package_id)
        return cpu_ids

    def __init__(self):
        super().__init__()
        self.cpu_ids = self.__get_cpu_ids()


class PowerGadgetLinuxRAPL(PowerGadgetLinux):
    """
    RAPL Linux custom PowerGadget wrapper.
    """

    @staticmethod
    def __get_cpu_energy(cpu):
        """
        Returns the CPU Energy figure
        """
        cpu_energy_file = Path(READ_RAPL_PATH.format(cpu)) / RAPL_ENERGY_FILE
        energy = int(cpu_energy_file.read_text())
        return energy

    @staticmethod
    def __get_dram_energy(cpu, dram):
        """
        Returns the dram energy figure
        """
        dram_energy_file = (
            Path(READ_RAPL_PATH.format(cpu))
            / (RAPL_DRAM_PATH.format(cpu, dram))
            / RAPL_ENERGY_FILE
        )
        energy = int(dram_energy_file.read_text())
        return energy

    def __init__(self):
        super().__init__()
        self.dram_ids = self.__get_drams_ids()
        self.power_draws = []
        self.record = {}
        self.start_time = None

    def __get_drams_ids(self):
        """
        Returns the drams id of this machine
        """
        dram_id_list = []
        for cpu in self.cpu_ids:
            dram_paths = glob.glob(
                READ_RAPL_PATH.format(cpu)
                + RAPL_DRAM_PATH.format(cpu, "*")
                + RAPL_DEVICENAME_FILE
            )
            for i, dram_file in enumerate(dram_paths):
                dram_file = Path(dram_file)
                if "dram" in dram_file.read_text():
                    dram_id_list.append((cpu, i))
                    break
        return dram_id_list

    def collect_power_usage(self):
        # ! not tested
        usages = pd.DataFrame(self.power_draws)
        usages[[TOTAL_ENERGY_CPU, TOTAL_ENERGY_MEMORY]] = (
            usages[["energy_cpu", "energy_memory"]]
            .diff()
            .fillna(usages[["energy_cpu", "energy_memory"]])
        )
        usages[TOTAL_ENERGY_PROCESS_CPU] = (
            usages[TOTAL_ENERGY_CPU] * usages["cpu_usage"]
        )
        usages[TOTAL_ENERGY_PROCESS_MEMORY] = (
            usages[TOTAL_ENERGY_MEMORY] * usages["memory_usage"]
        )
        return usages

    def __append_energy_usage(self, process, interval=1):
        # ! not tested
        energy_usage = {}
        energy_usage["energy_cpu"] = sum(
            [self.__get_cpu_energy(cpu_id) for cpu_id in self.cpu_ids]
        )
        energy_usage["energy_memory"] = sum(
            [
                self.__get_dram_energy(cpu_id, dram_id)
                for cpu_id, dram_id in self.dram_ids
            ]
        )
        _, cpu_usage, memory_usage = self.get_computer_usage(
            process, interval=interval
        )
        energy_usage["cpu_usage"] = cpu_usage
        energy_usage["memory_usage"] = memory_usage
        self.power_draws.append(energy_usage)

    def get_power_consumption(self, interval=1):
        # ! not tested
        """

        Parameters
        ----------
        interval (int)
        """
        current_process = psutil.Process()
        self.__append_energy_usage(current_process, interval=interval)
        while getattr(self.thread, "do_run", True):
            self.__append_energy_usage(current_process, interval=interval)
        self.__append_energy_usage(current_process, interval=interval)

    def start(self):
        LOGGER.info("starting CPU power monitoring ...")
        self.start_time = time.time()
        self.power_draws = []
        self.record = {}
        if self.thread and self.thread.is_alive():
            self.stop_thread()
        self.thread = threading.Thread(
            target=self.get_power_consumption, args=()
        )
        self.thread.start()

    def stop(self):
        LOGGER.info("stoping CPU power monitoring ...")
        usages = self.collect_power_usage()
        end_time = time.time()
        self.record[TOTAL_ENERGY_CPU] = (
            usages[TOTAL_ENERGY_CPU].sum() / 3600 / 1000
        )
        self.record[TOTAL_ENERGY_PROCESS_CPU] = (
            usages[TOTAL_ENERGY_PROCESS_CPU].sum() / 3600 / 1000
        )
        self.record[TOTAL_ENERGY_PROCESS_MEMORY] = (
            usages[TOTAL_ENERGY_PROCESS_MEMORY].sum() / 3600 / 1000
        )
        self.record[TOTAL_ENERGY_MEMORY] = (
            usages[TOTAL_ENERGY_MEMORY].sum() / 3600 / 1000
        )
        self.record[TOTAL_CPU_TIME] = end_time - self.start_time
        self.record[TOTAL_ENERGY_ALL] = 0


class PowerGadgetLinuxMSR(PowerGadgetLinux):
    """
    Mac OS X custom PowerGadget wrapper.
    """

    MSR_RAPL_POWER_UNIT = 0x606
    MSR_PKG_RAPL_POWER_LIMIT = 0x610
    # MSR_PKG_ENERGY_STATUS reports measured actual energy usage
    MSR_PKG_ENERGY_STATUS = 0x611
    MSR_DRAM_ENERGY_STATUS = 0x619
    MSR_PKG_PERF_STATUS = 0x613
    MSR_PKG_POWER_INFO = 0x614
    MSR_PP0_ENERGY_STATUS = 0x639

    @staticmethod
    def __read_msr(msr_file, msr_position):
        """
        Read MSR file at a specific position

        Parameters
        ----------
        msr_file : _io.TextIOWrapper
            An opened MSR file
        msr_position : int
            Position at which read the MSR file

        Returns
        -------
        int
            The value of the given MSR
        """
        os.lseek(msr_file, msr_position, os.SEEK_SET)
        return struct.unpack("Q", os.read(msr_file, 8))[0]

    def __init__(self):
        super().__init__()
        # the user needs to execute as root
        if os.getuid() != 0:
            raise PermissionError("You need to execute this program as root")
        self.thread = None
        self.power_draws = {}

    def __get_used_units(self, cpu):
        """
        Get the unit used by the MSR to encode the energy usage
        """
        msr_file = os.open(READ_MSR_PATH.format(cpu), os.O_RDONLY)
        # Calculate the units used
        result = self.__read_msr(msr_file, self.MSR_RAPL_POWER_UNIT)
        power_units = 0.5 ** (result & 0xF)
        cpu_energy_units = 0.5 ** ((result >> 8) & 0x1F)
        dram_energy_units = cpu_energy_units
        time_units = 0.5 ** ((result >> 16) & 0xF)
        os.close(msr_file)
        return power_units, cpu_energy_units, dram_energy_units, time_units

    def __get_cpu_energy(self, cpu, unit):
        """
        Returns the CPU Energy figure
        """
        msr_file = os.open(READ_MSR_PATH.format(cpu), os.O_RDONLY)
        result = self.__read_msr(msr_file, self.MSR_PKG_ENERGY_STATUS)
        os.close(msr_file)
        return result * unit / 3.6

    def __get_dram_energy(self, cpu, unit):
        """
        Returns the dram energy figure
        """
        msr_file = os.open(READ_MSR_PATH.format(cpu), os.O_RDONLY)
        result = self.__read_msr(msr_file, self.MSR_DRAM_ENERGY_STATUS)
        os.close(msr_file)
        return result * unit / 3.6

    def __get_computer_consumption(
        self, prev_cpu_energies, prev_dram_energies
    ):
        """
        Get the power consumption since the last measure

        Parameters
        ----------
        prev_cpu_energies : List
            List of the previous measure on the CPUs
            (One element of the list for each CPU)
        prev_dram_energies : [type]
            List of the previous measure on the DRAMs
            (One element of the list for each CPU)

        Returns
        -------
        [type]
            [description]
        """
        cpu_power = 0
        dram_power = 0
        for i, cpu in enumerate(self.cpu_ids):
            _, cpu_energy_units, dram_energy_units, _ = self.__get_used_units(
                cpu
            )
            current_dram_energy = self.__get_dram_energy(
                cpu, dram_energy_units
            )
            current_cpu_energy = self.__get_cpu_energy(cpu, cpu_energy_units)
            if prev_cpu_energies[i] > cpu_energy_units:
                cpu_energy_units *= 2
            if prev_dram_energies[i] > dram_energy_units:
                dram_energy_units *= 2
            cpu_power += current_cpu_energy - prev_cpu_energies[i]
            dram_power += current_dram_energy - prev_dram_energies[i]
            prev_cpu_energies[i] = current_cpu_energy
            prev_dram_energies[i] = current_dram_energy
        return cpu_power, dram_power, prev_cpu_energies, prev_dram_energies

    def get_computer_consumption(self, interval=1):
        """

        Parameters
        ----------
        interval (int)
        """
        # ! not tested
        process = psutil.Process()
        self.power_draws[TOTAL_CPU_TIME] = 0
        self.power_draws[TOTAL_ENERGY_CPU] = 0
        self.power_draws[TOTAL_ENERGY_MEMORY] = 0
        prev_dram_energies = []
        prev_cpu_energies = []
        for cpu in self.cpu_ids:
            _, cpu_energy_units, dram_energy_units, _ = self.__get_used_units(
                cpu
            )
            prev_dram_energies.append(
                self.__get_dram_energy(cpu, dram_energy_units)
            )
            prev_cpu_energies.append(
                self.__get_cpu_energy(cpu, cpu_energy_units)
            )
        start_time = time.time()
        while getattr(self.thread, "do_run", True):
            # time.sleep(interval)
            _, cpu_usage, memory_usage = self.get_computer_usage(
                process, interval=interval
            )
            (
                cpu_power,
                dram_power,
                prev_cpu_energies,
                prev_dram_energies,
            ) = self.__get_computer_consumption(
                prev_cpu_energies, prev_dram_energies
            )
            self.power_draws[TOTAL_ENERGY_CPU] += cpu_power
            self.power_draws[TOTAL_ENERGY_MEMORY] += dram_power
            self.power_draws[TOTAL_ENERGY_PROCESS_CPU] += cpu_power * cpu_usage
            self.power_draws[TOTAL_ENERGY_PROCESS_MEMORY] += (
                dram_power * memory_usage
            )
            end_time = time.time()
            self.power_draws[TOTAL_CPU_TIME] += end_time - start_time
            start_time = end_time

    def start(self):
        LOGGER.info("starting CPU power monitoring ...")
        if self.thread and self.thread.is_alive():
            self.stop()
        self.power_draws = {}
        self.thread = threading.Thread(
            target=self.get_computer_consumption, args=()
        )
        self.thread.start()

    def stop(self):
        LOGGER.info("stoping CPU power monitoring ...")
        self.thread.do_run = False
        self.thread.join()
        self.record = self.power_draws
        self.record[TOTAL_ENERGY_ALL] = 0
